from xdsl.context import Context
from xdsl.dialects import arith, builtin, scf
from xdsl.ir import SSAValue
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)


class UnrollLoopPattern(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: scf.ForOp, rewriter: PatternRewriter) -> None:
        if (
            not isinstance(lb_op := op.lb.owner, arith.ConstantOp)
            or not isinstance(ub_op := op.ub.owner, arith.ConstantOp)
            or not isinstance(step_op := op.step.owner, arith.ConstantOp)
        ):
            return

        assert isinstance(lb_op.value, builtin.IntegerAttr)
        assert isinstance(ub_op.value, builtin.IntegerAttr)
        assert isinstance(step_op.value, builtin.IntegerAttr)

        lb = lb_op.value.value.data
        ub = ub_op.value.value.data
        step = step_op.value.value.data

        iter_args: tuple[SSAValue, ...] = op.iter_args

        i_arg, *block_iter_args = op.body.block.args

        for i in range(lb, ub, step):
            i_op = rewriter.insert(
                arith.ConstantOp(builtin.IntegerAttr(i, lb_op.value.type))
            )
            i_op.result.name_hint = i_arg.name_hint

            value_mapper: dict[SSAValue, SSAValue] = {
                arg: val for arg, val in zip(block_iter_args, iter_args, strict=True)
            }
            value_mapper[i_arg] = i_op.result

            for inner_op in op.body.block.ops:
                if isinstance(inner_op, scf.YieldOp):
                    iter_args = tuple(
                        value_mapper.get(val, val) for val in inner_op.arguments
                    )
                else:
                    rewriter.insert(inner_op.clone(value_mapper))

        rewriter.replace_matched_op((), iter_args)


class ScfForLoopUnrollPass(ModulePass):
    """
    Fully unrolls all loops where the lb, ub, and step are constants.
    """

    name = "scf-for-loop-unroll"

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        PatternRewriteWalker(UnrollLoopPattern()).rewrite_module(op)



from xdsl.context import Context
from xdsl.dialects import builtin
from xdsl.frontend.listlang import list_dialect
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.rewriter import InsertPoint
from xdsl.utils.hints import isa


class LengthOfMap(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: list_dialect.LengthOp, rewriter: PatternRewriter):
        if not isa(op.li.owner, list_dialect.MapOp):
            return

        rewriter.replace_op(op, list_dialect.LengthOp(op.li.owner.li))


class MapOfMap(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: list_dialect.MapOp, rewriter: PatternRewriter):
        if not isa(op.li.owner, list_dialect.MapOp):
            return

        if op.li.uses.get_length() != 1:
            return

        first_block = op.li.owner.body.block
        second_block = op.body.block

        assert isa(first_block.last_op, list_dialect.YieldOp)
        first_block_result = first_block.last_op.yielded
        rewriter.erase_op(first_block.last_op)

        rewriter.inline_block(
            second_block, InsertPoint.at_end(first_block), [first_block_result]
        )

        rewriter.replace_value_with_new_type(op.li.owner.result, op.result.type)

        rewriter.replace_all_uses_with(op.result, op.li)
        rewriter.erase_op(op)


class OptimizeListOps(ModulePass):
    """
    Applies optimizations to list operations.
    """

    name = "optimize-lists"

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    LengthOfMap(),
                    MapOfMap(),
                ]
            ),
        ).rewrite_module(op)



from collections.abc import Sequence
from dataclasses import dataclass

import xdsl.frontend.listlang.list_dialect as list_dialect
from xdsl.builder import Builder
from xdsl.dialects import builtin, printf
from xdsl.frontend.listlang.source import Located, ParseError
from xdsl.ir import Attribute, Block, Region, SSAValue
from xdsl.utils.hints import isa


class ListLangType:
    @staticmethod
    def from_xdsl(xdsl_type: Attribute) -> "ListLangType":
        if isa(xdsl_type, builtin.IntegerType[32]):
            return ListLangInt()
        if isa(xdsl_type, builtin.IntegerType[1]):
            return ListLangBool()
        if isa(xdsl_type, builtin.TensorType[builtin.IntegerType[32]]):
            return ListLangList(ListLangInt())
        if isa(xdsl_type, builtin.TensorType[builtin.IntegerType[1]]):
            return ListLangList(ListLangBool())
        raise ValueError("unknown type")

    def __str__(self) -> str: ...
    def xdsl(self) -> Attribute: ...
    def print(self, builder: Builder, value: SSAValue): ...
    def get_method(self, method: str) -> "Method | None":
        return None


@dataclass
class ListLangInt(ListLangType):
    def __str__(self) -> str:
        return "int"

    def xdsl(self) -> builtin.IntegerType:
        return builtin.IntegerType(32)

    def print(self, builder: Builder, value: SSAValue):
        builder.insert_op(printf.PrintFormatOp("{}", value))


@dataclass
class ListLangBool(ListLangType):
    def __str__(self) -> str:
        return "bool"

    def xdsl(self) -> builtin.IntegerType:
        return builtin.IntegerType(1)

    def print(self, builder: Builder, value: SSAValue):
        builder.insert_op(printf.PrintFormatOp("{}", value))


LIST_ELEMENT_TYPE = ListLangBool | ListLangInt


@dataclass
class ListLangList(ListLangType):
    element_type: LIST_ELEMENT_TYPE

    def __str__(self) -> str:
        return f"list<{self.element_type}>"

    def xdsl(self) -> list_dialect.ListType:
        return list_dialect.ListType(self.element_type.xdsl())

    def print(self, builder: Builder, value: SSAValue):
        builder.insert_op(list_dialect.PrintOp(value))

    def get_method(self, method: str) -> "Method | None":
        match method:
            case "len":
                return ListLenMethod()
            case "map":
                return ListMapMethod()
            case _:
                return None


@dataclass
class TypedExpression:
    value: SSAValue
    typ: ListLangType


## Methods


class Method:
    name: str

    def get_lambda_arg_type(self, x: ListLangType) -> Sequence[ListLangType] | None:
        """
        From the type on which the method was invoked, returns the types of the
        arguments of the method's lambda if there is one, or None if there is
        no lambda.
        """
        ...

    def build(
        self,
        builder: Builder,
        x: Located[TypedExpression],
        lambd: Located[tuple[Block, ListLangType]] | None,
    ) -> TypedExpression:
        """
        Builds the method's execution.

        `lambd` contains a free-standing block containing the lambda
        instructions that must be inlined as needed, and the type of the final
        expression of the block. The associated location is the location of
        the result expression.
        """
        ...


class ListLenMethod(Method):
    name = "len"

    def get_lambda_arg_type(self, x: ListLangType) -> Sequence[ListLangType] | None:
        return None

    def build(
        self,
        builder: Builder,
        x: Located[TypedExpression],
        lambd: Located[tuple[Block, ListLangType]] | None,
    ) -> TypedExpression:
        assert lambd is None
        assert isinstance(x.value.typ, ListLangList)

        len_op = builder.insert_op(list_dialect.LengthOp(x.value.value))
        return TypedExpression(len_op.result, ListLangInt())


class ListMapMethod(Method):
    name = "map"

    def get_lambda_arg_type(self, x: ListLangType) -> Sequence[ListLangType] | None:
        assert isinstance(x, ListLangList)
        return (x.element_type,)

    def build(
        self,
        builder: Builder,
        x: Located[TypedExpression],
        lambd: Located[tuple[Block, ListLangType]] | None,
    ) -> TypedExpression:
        assert lambd is not None
        assert isinstance(x.value.typ, ListLangList)

        if not isinstance(lambd.value[1], LIST_ELEMENT_TYPE):
            raise ParseError(lambd.loc.pos, "expected expression of list-element type")

        map_op = builder.insert_op(
            list_dialect.MapOp(
                x.value.value, Region([lambd.value[0]]), lambd.value[1].xdsl()
            )
        )

        return TypedExpression(map_op.result, ListLangList(lambd.value[1]))



import re
from typing import Literal

import pytest
from typing_extensions import TypeVar

from xdsl.irdl import (
    AnyInt,
    AtLeast,
    AtMost,
    ConstraintContext,
    EqIntConstraint,
    IntSetConstraint,
    IntTypeVarConstraint,
    get_int_constraint,
    get_optional_int_constraint,
)
from xdsl.utils.exceptions import PyRDLTypeError, VerifyException


def test_failing_inference():
    with pytest.raises(
        ValueError,
        match=re.escape(r"Cannot infer integer from constraint AtLeast(bound=3)"),
    ):
        AtLeast(3).infer(ConstraintContext())


def test_at_least():
    AtLeast(0).verify(0, ConstraintContext())

    AtLeast(1).verify(1, ConstraintContext())
    AtLeast(1).verify(2, ConstraintContext())
    with pytest.raises(VerifyException):
        AtLeast(1).verify(0, ConstraintContext())

    AtLeast(2).verify(2, ConstraintContext())
    AtLeast(2).verify(3, ConstraintContext())
    with pytest.raises(VerifyException):
        AtLeast(2).verify(1, ConstraintContext())
    with pytest.raises(VerifyException):
        AtLeast(2).verify(0, ConstraintContext())


def test_at_most():
    AtMost(0).verify(0, ConstraintContext())

    AtMost(1).verify(1, ConstraintContext())
    AtMost(1).verify(0, ConstraintContext())
    with pytest.raises(VerifyException, match="Expected integer <= 0, got 1"):
        AtMost(0).verify(1, ConstraintContext())

    AtMost(2).verify(2, ConstraintContext())
    AtMost(2).verify(1, ConstraintContext())
    with pytest.raises(VerifyException, match="Expected integer <= 2, got 3"):
        AtMost(2).verify(3, ConstraintContext())


def test_eq():
    one_constr = EqIntConstraint(1)
    one_constr.verify(1, ConstraintContext())
    with pytest.raises(VerifyException, match="Invalid value 2, expected 1"):
        one_constr.verify(2, ConstraintContext())
    assert one_constr.can_infer(set())
    assert one_constr.infer(ConstraintContext()) == 1


def test_set():
    empty_constr = IntSetConstraint(frozenset())
    assert not empty_constr.can_infer(set())
    with pytest.raises(
        VerifyException, match=re.escape("Invalid value 2, expected one of {}")
    ):
        empty_constr.verify(2, ConstraintContext())

    one_constr = IntSetConstraint(frozenset((0,)))
    one_constr.verify(0, ConstraintContext())
    assert one_constr.can_infer(set())
    assert one_constr.infer(ConstraintContext()) == 0
    with pytest.raises(
        VerifyException, match=re.escape("Invalid value 2, expected one of {0}")
    ):
        one_constr.verify(2, ConstraintContext())

    two_constr = IntSetConstraint(frozenset((0, 1)))
    two_constr.verify(0, ConstraintContext())
    assert not two_constr.can_infer(set())
    with pytest.raises(
        VerifyException, match=re.escape("Invalid value 2, expected one of {0, 1}")
    ):
        two_constr.verify(2, ConstraintContext())


def test_mapping_type_vars():
    _IntT = TypeVar("_IntT", bound=int, default=int)
    tv_constr = IntTypeVarConstraint(_IntT, AnyInt())
    my_constr = EqIntConstraint(1)
    assert tv_constr.mapping_type_vars({_IntT: my_constr}) is my_constr


def test_get_int_constr():
    assert get_int_constraint(int) == AnyInt()
    assert get_int_constraint(Literal[1]) == EqIntConstraint(1)
    assert get_int_constraint(Literal[2]) == EqIntConstraint(2)
    assert get_int_constraint(Literal[2, 3]) == IntSetConstraint(frozenset((2, 3)))

    assert get_int_constraint(
        Literal[2] | Literal[3]  # noqa
    ) == IntSetConstraint(frozenset((2, 3)))

    assert get_int_constraint(
        Literal[1] | Literal[2, 3]  # noqa
    ) == IntSetConstraint(frozenset((1, 2, 3)))
    assert get_int_constraint(
        (Literal[1] | Literal[2]) | Literal[3]  # noqa
    ) == IntSetConstraint(frozenset((1, 2, 3)))
    assert get_int_constraint(
        Literal[1] | (Literal[2] | Literal[3])  # noqa
    ) == IntSetConstraint(frozenset((1, 2, 3)))

    assert get_optional_int_constraint(str) is None

    with pytest.raises(PyRDLTypeError, match="Unexpected int type: <class 'str'>"):
        get_int_constraint(str)  # pyright: ignore[reportArgumentType]



from __future__ import annotations

from dataclasses import dataclass
from unittest.mock import Mock

import pytest

from xdsl.analysis.dataflow import (
    AnalysisState,
    ChangeResult,
    DataFlowAnalysis,
    DataFlowSolver,
    GenericLatticeAnchor,
    LatticeAnchor,
    ProgramPoint,
)
from xdsl.context import Context
from xdsl.dialects import test
from xdsl.dialects.builtin import IntegerType, ModuleOp
from xdsl.ir import Block, Operation, Region
from xdsl.utils.test_value import create_ssa_value


# region ChangeResult tests
def test_change_result_or():
    assert ChangeResult.CHANGE | ChangeResult.CHANGE is ChangeResult.CHANGE
    assert ChangeResult.CHANGE | ChangeResult.NO_CHANGE is ChangeResult.CHANGE
    assert ChangeResult.NO_CHANGE | ChangeResult.CHANGE is ChangeResult.CHANGE
    assert ChangeResult.NO_CHANGE | ChangeResult.NO_CHANGE is ChangeResult.NO_CHANGE


# endregion


# region GenericLatticeAnchor tests
@dataclass(frozen=True)
class MyAnchor(GenericLatticeAnchor):
    name: str

    def __str__(self) -> str:
        return f"MyAnchor({self.name})"


def test_generic_lattice_anchor():
    anchor1 = MyAnchor("a")
    anchor2 = MyAnchor("a")
    anchor3 = MyAnchor("b")

    assert anchor1 == anchor2
    assert anchor1 != anchor3
    assert hash(anchor1) == hash(anchor2)
    assert hash(anchor1) != hash(anchor3)
    assert str(anchor1) == "MyAnchor(a)"
    assert not (anchor1 == "a")


# endregion


# region ProgramPoint tests
@pytest.fixture
def ops_in_block() -> tuple[Operation, Operation, Operation, Block]:
    context = Context()
    context.allow_unregistered = True
    op1_class = context.get_optional_op("test_op1")
    op2_class = context.get_optional_op("test_op2")
    op3_class = context.get_optional_op("test_op3")
    assert op1_class is not None
    assert op2_class is not None
    assert op3_class is not None
    op1 = op1_class()
    op2 = op2_class()
    op3 = op3_class()
    block = Block([op1, op2, op3])
    return op1, op2, op3, block


@pytest.fixture
def empty_block() -> Block:
    return Block()


def test_program_point_before(
    ops_in_block: tuple[Operation, Operation, Operation, Block],
):
    _, op2, _, _ = ops_in_block
    pp = ProgramPoint.before(op2)
    assert pp.entity is op2


def test_program_point_after(
    ops_in_block: tuple[Operation, Operation, Operation, Block],
):
    op1, op2, op3, block = ops_in_block

    # After op in middle of block
    pp_after_op1 = ProgramPoint.after(op1)
    assert pp_after_op1.entity is op2

    # After last op in block
    pp_after_op3 = ProgramPoint.after(op3)
    assert pp_after_op3.entity is block


def test_program_point_after_detached():
    detached_op = test.TestOp()
    with pytest.raises(
        ValueError, match="Cannot get ProgramPoint after a detached operation."
    ):
        ProgramPoint.after(detached_op)


def test_program_point_at_block_boundaries(
    ops_in_block: tuple[Operation, Operation, Operation, Block], empty_block: Block
):
    op1, _, _, block = ops_in_block

    # Start of non-empty block
    pp_start = ProgramPoint.at_start_of_block(block)
    assert pp_start.entity is op1

    # End of non-empty block
    pp_end = ProgramPoint.at_end_of_block(block)
    assert pp_end.entity is block

    # Start of empty block
    pp_start_empty = ProgramPoint.at_start_of_block(empty_block)
    assert pp_start_empty.entity is empty_block

    # End of empty block
    pp_end_empty = ProgramPoint.at_end_of_block(empty_block)
    assert pp_end_empty.entity is empty_block


def test_program_point_properties(
    ops_in_block: tuple[Operation, Operation, Operation, Block],
):
    op1, _, _, block = ops_in_block

    pp_before_op = ProgramPoint.before(op1)
    assert pp_before_op.op is op1
    assert pp_before_op.block is block

    pp_at_end = ProgramPoint.at_end_of_block(block)
    assert pp_at_end.op is None
    assert pp_at_end.block is block


# endregion


# region AnalysisState tests
class MyState(AnalysisState):
    def __str__(self) -> str:
        return "MyState"


def test_analysis_state_on_update():
    solver = Mock(spec=DataFlowSolver)
    anchor = create_ssa_value(IntegerType(32))
    state = MyState(anchor)

    point1 = ProgramPoint.before(test.TestOp())
    analysis1 = Mock(spec=DataFlowAnalysis)

    point2 = ProgramPoint.at_end_of_block(Block())
    analysis2 = Mock(spec=DataFlowAnalysis)

    state.dependents.add((point1, analysis1))
    state.dependents.add((point2, analysis2))

    state.on_update(solver)

    solver.enqueue.assert_any_call((point1, analysis1))
    solver.enqueue.assert_any_call((point2, analysis2))
    assert solver.enqueue.call_count == 2


# endregion


# region DataFlowSolver tests
@pytest.fixture
def solver() -> DataFlowSolver:
    return DataFlowSolver(Context())


class MyAnalysis(DataFlowAnalysis):
    def initialize(self, op: Operation) -> None:
        pass

    def visit(self, point: ProgramPoint) -> None:
        pass


def test_data_flow_solver_init(solver: DataFlowSolver):
    # solver should be able to load analyses
    # and accept basic operations without errors
    assert isinstance(solver.context, Context)
    analysis = solver.load(MyAnalysis)
    assert isinstance(analysis, MyAnalysis)


def test_data_flow_solver_load(solver: DataFlowSolver):
    analysis = solver.load(MyAnalysis)
    assert isinstance(analysis, MyAnalysis)
    assert analysis.solver is solver


def test_data_flow_solver_load_while_running_raises(solver: DataFlowSolver):
    # Create an analysis that tries to load another analysis during execution
    class LoadingAnalysis(DataFlowAnalysis):
        def initialize(self, op: Operation) -> None:
            # Try to load while running
            with pytest.raises(
                RuntimeError,
                match="Cannot load new analyses while the solver is running.",
            ):
                self.solver.load(MyAnalysis)

        def visit(self, point: ProgramPoint) -> None:
            pass

    solver.load(LoadingAnalysis)
    solver.initialize_and_run(Mock())


def test_data_flow_solver_get_lookup_state(solver: DataFlowSolver):
    anchor = create_ssa_value(IntegerType(32))

    # Test get_or_create_state
    state1 = solver.get_or_create_state(anchor, MyState)
    assert isinstance(state1, MyState)
    assert state1.anchor is anchor

    # Test that it returns the same state
    state2 = solver.get_or_create_state(anchor, MyState)
    assert state1 is state2

    # Test lookup_state
    state3 = solver.lookup_state(anchor, MyState)
    assert state1 is state3

    # Test lookup_state for non-existent state
    anchor2 = create_ssa_value(IntegerType(64))
    assert solver.lookup_state(anchor2, MyState) is None
    assert solver.lookup_state(anchor, AnalysisState) is None


def test_data_flow_solver_enqueue_not_running(solver: DataFlowSolver):
    # This tests the public contract - enqueue should only work while running
    with pytest.raises(
        RuntimeError, match="Cannot enqueue work items when the solver is not running."
    ):
        solver.enqueue((Mock(), Mock()))


def test_data_flow_solver_propagate_not_running(solver: DataFlowSolver):
    # This tests the public contract - propagate should only work while running
    with pytest.raises(
        RuntimeError, match="Cannot propagate changes when the solver is not running."
    ):
        solver.propagate_if_changed(Mock(), ChangeResult.CHANGE)


def test_data_flow_solver_run_twice_raises(solver: DataFlowSolver):
    # Create an analysis that tries to run the solver again during execution
    class ReentrantAnalysis(DataFlowAnalysis):
        def initialize(self, op: Operation) -> None:
            # Try to run again while already running
            with pytest.raises(RuntimeError, match="Solver is already running."):
                self.solver.initialize_and_run(Mock())

        def visit(self, point: ProgramPoint) -> None:
            pass

    solver.load(ReentrantAnalysis)
    solver.initialize_and_run(Mock())


# State for the test
class CounterState(AnalysisState):
    def __init__(self, anchor: LatticeAnchor):
        super().__init__(anchor)
        self.count = 0

    def __str__(self) -> str:
        return f"CounterState({self.count})"


# Analysis for the test
class CounterAnalysis(DataFlowAnalysis):
    def initialize(self, op: Operation) -> None:
        # Enqueue the start of the first block
        start_point = ProgramPoint.at_start_of_block(op.regions[0].blocks[0])
        self.solver.enqueue((start_point, self))

    def visit(self, point: ProgramPoint) -> None:
        state = self.get_or_create_state(point, CounterState)

        # Only visit each point twice
        if state.count >= 2:
            return

        state.count += 1

        # Add a dependency on self to re-trigger the visit
        self.add_dependency(state, point)

        self.propagate_if_changed(state, ChangeResult.CHANGE)


def test_data_flow_solver_run_loop(solver: DataFlowSolver):
    op1 = test.TestOp()
    block = Block([op1])

    module = ModuleOp([test.TestOp(regions=[Region([block])])])

    assert module.body.first_block is not None
    assert module.body.first_block.first_op is not None
    solver.load(CounterAnalysis)
    solver.initialize_and_run(module.body.first_block.first_op)

    start_point = ProgramPoint.at_start_of_block(block)
    state = solver.lookup_state(start_point, CounterState)

    assert state is not None
    assert state.count == 2


# endregion


# region DataFlowAnalysis tests
def test_data_flow_analysis_helpers(solver: DataFlowSolver):
    analysis = MyAnalysis(solver)

    # Mock the solver methods to check if they are called
    solver.get_or_create_state = Mock(return_value="state")
    solver.lookup_state = Mock(return_value="state")
    solver.propagate_if_changed = Mock()

    anchor = create_ssa_value(IntegerType(32))

    # test get_or_create_state
    res = analysis.get_or_create_state(anchor, MyState)
    solver.get_or_create_state.assert_called_once_with(anchor, MyState)
    assert res == "state"

    # test get_state
    res = analysis.get_state(anchor, MyState)
    solver.lookup_state.assert_called_once_with(anchor, MyState)
    assert res == "state"

    # test add_dependency
    state = MyState(anchor)
    point = ProgramPoint.before(test.TestOp())
    analysis.add_dependency(state, point)
    assert (point, analysis) in state.dependents

    # test propagate_if_changed
    analysis.propagate_if_changed(state, ChangeResult.CHANGE)
    solver.propagate_if_changed.assert_called_once_with(state, ChangeResult.CHANGE)


# endregion



import json
import os
from dataclasses import dataclass, field

from xdsl.context import Context
from xdsl.dialects import builtin, eqsat
from xdsl.dialects.builtin import IntAttr
from xdsl.ir import Block, Operation, OpResult, SSAValue
from xdsl.passes import ModulePass
from xdsl.utils.exceptions import DiagnosticException
from xdsl.utils.hints import isa


def get_node_base_cost(op: Operation, default_cost: int | None) -> int | None:
    """
    Get the base cost of an operation (without considering dependencies).
    """
    cost_attr = op.attributes.get(eqsat.EQSAT_COST_LABEL)
    if cost_attr is None:
        return default_cost
    if not isa(cost_attr, IntAttr):
        raise DiagnosticException(
            f"Unexpected value {cost_attr} for key {eqsat.EQSAT_COST_LABEL} in {op}"
        )
    return cost_attr.data


def calculate_node_total_cost(
    value: SSAValue, eclass_costs: dict[OpResult, int], default_cost: int | None
) -> int | None:
    """
    Calculate the total cost of a node: its own cost plus the costs of all its
    e-class dependencies. This is equivalent to the Rust `node_sum_cost` function.

    Uses dictionary lookup (not recursion) to get child e-class costs.
    Returns None if any dependency cost is unknown.
    """
    # Block arguments are free
    if not isinstance(value, OpResult):
        return 0

    op = value.op

    # For e-classes, return their current best known cost (None if not set)
    if isinstance(op, eqsat.AnyEClassOp):
        return eclass_costs.get(value)

    # For regular operations, compute: own cost + sum of dependent e-class costs
    node_cost = get_node_base_cost(op, default_cost)
    if node_cost is None:
        return None

    total = node_cost

    # Add costs of all operands (non-recursive, just dictionary lookup)
    for operand in op.operands:
        if isinstance(operand, OpResult) and isinstance(operand.op, eqsat.AnyEClassOp):
            # Look up the e-class cost from the dictionary
            operand_cost = eclass_costs.get(operand)
            if operand_cost is None:
                return None
        else:
            # Block argument or non-eclass operation
            operand_cost = (
                0
                if not isinstance(operand, OpResult)
                else get_node_base_cost(operand.op, default_cost) or 0
            )

        total += operand_cost

    return total


def add_eqsat_costs(block: Block, default: int | None, cost_dict: dict[str, int]):
    """
    Add costs to all operations and perform bottom-up extraction to find the minimum
    cost node in each e-class using fixed-point iteration.
    """
    # First pass: assign base costs to operations from cost_dict or default
    for op in block.ops:
        if not op.results:
            continue

        if eqsat.EQSAT_COST_LABEL in op.attributes:
            continue

        if op.name in cost_dict:
            op.attributes[eqsat.EQSAT_COST_LABEL] = IntAttr(cost_dict[op.name])
            continue

        if len(op.results) != 1:
            raise DiagnosticException(
                "Cannot compute cost of one result of operation with multiple "
                f"results: {op}"
            )

        # For non-eclass operations without explicit costs, use default
        if not isinstance(op, eqsat.AnyEClassOp):
            if default is not None:
                op.attributes[eqsat.EQSAT_COST_LABEL] = IntAttr(default)

    # Track the minimum total cost for each e-class
    eclass_costs: dict[OpResult, int] = {}

    changed = True
    while changed:
        changed = False

        # Process all e-class operations
        for op in block.ops:
            if not isinstance(op, eqsat.AnyEClassOp):
                continue

            if not op.results:
                continue

            eclass_result = op.results[0]

            # For each operand (node) in this e-class, calculate its total cost
            for idx, operand in enumerate(op.operands):
                total_cost = calculate_node_total_cost(operand, eclass_costs, default)

                # Skip if cost cannot be determined yet
                if total_cost is None:
                    continue

                # Get current best cost for this e-class (None if not set)
                current_best = eclass_costs.get(eclass_result)

                # Update if this operand has lower cost (or if no cost is set yet)
                if current_best is None or total_cost < current_best:
                    eclass_costs[eclass_result] = total_cost
                    op.min_cost_index = IntAttr(idx)
                    changed = True


@dataclass(frozen=True)
class EqsatAddCostsPass(ModulePass):
    """
    Add costs to all operations in blocks that contain eqsat.eclass ops, and perform
    bottom-up extraction to find the minimum cost node in each e-class.

    The cost of an eclass operation is determined through fixed-point iteration:
    - Each operand's total cost is calculated (own cost + dependency costs)
    - The operand with minimum total cost is selected and stored in min_cost_index

    The cost for non-eclass operations is fetched from the cost_dict or set to the
    default value. The cost is stored as an IntAttr in the EQSAT_COST_LABEL attribute.

    If the cost cannot be calculated, the default value can be provided with the
    `default` optional parameter.
    """

    name = "eqsat-add-costs"

    cost_file: str | None = field(default=None)
    "Path to JSON file of cost values"
    default: int | None = field(default=None)
    "Default cost to assign if it cannot be calculated."

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        eclass_parent_blocks = set(
            o.parent
            for o in op.walk()
            if o.parent is not None and isinstance(o, eqsat.AnyEClassOp)
        )

        cost_dict: dict[str, int] = {}

        if self.cost_file is not None:
            assert os.path.exists(self.cost_file)
            with open(self.cost_file) as file:
                cost_dict = json.load(file)

        for block in eclass_parent_blocks:
            add_eqsat_costs(block, default=self.default, cost_dict=cost_dict)



# RUN: python %s | filecheck %s

from xdsl.dialects import bigint
from xdsl.frontend.pyast.context import PyASTContext

ctx = PyASTContext(post_transforms=[])
ctx.register_type(int, bigint.bigint)


# CHECK:      func.func @f1(%{{.*}} : !bigint.bigint) {
# CHECK-NEXT:   symref.declare "x"
# CHECK-NEXT:   symref.update @x = %{{.*}} : !bigint.bigint
# CHECK-NEXT:   func.return
# CHECK-NEXT: }
@ctx.parse_program
def f1(x: int):
    return


print(f1.module)


# CHECK:      func.func @f2() {
# CHECK-NEXT:   func.return
# CHECK-NEXT: }
@ctx.parse_program
def f2():
    return


print(f2.module)


# CHECK:      func.func @f3(%{{.*}} : !bigint.bigint) -> !bigint.bigint {
# CHECK-NEXT:   symref.declare "x"
# CHECK-NEXT:   symref.update @x = %{{.*}} : !bigint.bigint
# CHECK-NEXT:   %{{.*}} = symref.fetch @x : !bigint.bigint
# CHECK-NEXT:   func.return %{{.*}} : !bigint.bigint
# CHECK-NEXT: }
@ctx.parse_program
def f3(x: int) -> int:
    return x


print(f3.module)

try:
    # CHECK:      func.func @f4(%{{.*}} : !bigint.bigint) -> !bigint.bigint {
    # CHECK-NEXT:   symref.declare "x"
    # CHECK-NEXT:   symref.update @x = %{{.*}} : !bigint.bigint
    # CHECK-NEXT:   %{{.*}} = symref.fetch @x : !bigint.bigint
    # CHECK-NEXT:   func.return %{{.*}} : !bigint.bigint
    # CHECK-NEXT: }
    @ctx.parse_program
    def f4(x: int) -> int:
        return x

    print(f4.module)
except Exception as e:
    print(e)
    exit(1)



import re

import pytest

from xdsl.ir import Dialect
from xdsl.universe import Universe


def test_multiverse():
    """
    Checks that the universe can be discovered within the current environment, and that
    duplicate names are detected.
    """
    multiverse = Universe.get_multiverse()

    assert "plugin_dialect" in multiverse.all_dialects
    assert "plugin-pass" in multiverse.all_passes

    import my_plugin

    my_plugin.MY_UNIVERSE.all_dialects["scf"] = lambda: Dialect("scf")

    with pytest.raises(
        ValueError,
        match=re.escape("Duplicate definition of scf in ['my_plugin', 'xdsl']."),
    ):
        Universe.get_multiverse()

    del my_plugin.MY_UNIVERSE.all_dialects["scf"]

    my_plugin.MY_UNIVERSE.all_passes["dce"] = lambda: my_plugin.PluginPass

    with pytest.raises(
        ValueError,
        match=re.escape("Duplicate definition of dce in ['my_plugin', 'xdsl']."),
    ):
        Universe.get_multiverse()

    del my_plugin.MY_UNIVERSE.all_passes["dce"]



from collections.abc import Sequence
from dataclasses import dataclass

from xdsl.builder import Builder
from xdsl.context import Context
from xdsl.dialect_interfaces.constant_materialization import (
    ConstantMaterializationInterface,
)
from xdsl.ir import Attribute, Operation, SSAValue, TypeAttribute
from xdsl.pattern_rewriter import PatternRewriter
from xdsl.traits import HasFolder


@dataclass
class Folder:
    context: Context

    def try_fold(self, op: Operation) -> tuple[list[SSAValue], list[Operation]] | None:
        """
        Try to fold the given operation.
        Returns a tuple with the list of SSAValues that replace the results of the operation,
        and a list of constant operations that were created from the constant attributes generated by the folder.
        If the operation could not be folded, returns None.

        Note that while this folds only one operation, multiple new operations can be created.
        Each of the results of the original operation might be replaced by a new constant operation.
        """

        if (trait := op.get_trait(HasFolder)) is None:
            return None
        folded = trait.fold(op)
        if folded is None:
            return None
        results: list[SSAValue] = []
        new_ops: list[Operation] = []
        for val, original_result in zip(folded, op.results, strict=True):
            if isinstance(val, SSAValue):
                results.append(val)
            else:
                assert isinstance(val, Attribute)
                dialect = self.context.get_dialect(op.dialect_name())
                interface = dialect.get_interface(ConstantMaterializationInterface)
                if interface is None:
                    return None
                assert isinstance(type := original_result.type, TypeAttribute)
                new_op = interface.materialize_constant(val, type)
                if new_op is None:
                    return None
                new_ops.append(new_op)
                results.append(new_op.results[0])
        return results, new_ops

    def insert_with_fold(
        self, op: Operation, builder: Builder
    ) -> Sequence[SSAValue] | None:
        """
        Inserts the operation using the provided builder, trying to fold it first.
        If folding is successful, the folded results are returned, otherwise None is returned.
        """
        results = self.try_fold(op)
        if results is None:
            builder.insert(op)
            return op.results
        else:
            if op.parent:
                raise ValueError(
                    "Can't insert_with_fold fold an operation that already has a parent."
                )
            values, new_ops = results
            builder.insert_op(new_ops)
            return values

    def replace_with_fold(
        self, op: Operation, rewriter: PatternRewriter, safe_erase: bool = True
    ) -> Sequence[SSAValue] | None:
        """
        Replaces the operation with its folded results.
        If folding is successful, the folded results are returned.
        Otherwise, returns None.
        """
        results = self.try_fold(op)
        if results is None:
            return None
        values, new_ops = results
        rewriter.replace_op(op, new_ops, values, safe_erase)
        return values



from __future__ import annotations

import abc
from typing import ClassVar, cast

from xdsl.dialects.arith import FastMathFlagsAttr
from xdsl.dialects.builtin import (
    AnyAttr,
    AnyFloat,
    AnyFloatConstr,
    ArrayAttr,
    ArrayOfConstraint,
    ComplexType,
    FloatAttr,
    IntegerAttr,
    IntegerType,
    ParamAttrConstraint,
)
from xdsl.interfaces import ConstantLikeInterface
from xdsl.ir import Attribute, Dialect, Operation, SSAValue
from xdsl.irdl import (
    AnyOf,
    EqIntConstraint,
    IRDLOperation,
    RangeOf,
    VarConstraint,
    base,
    irdl_op_definition,
    operand_def,
    prop_def,
    result_def,
    traits_def,
)
from xdsl.traits import Pure
from xdsl.utils.exceptions import VerifyException

ComplexTypeConstr = ComplexType.constr(AnyFloat)


class ComplexUnaryComplexResultOperation(IRDLOperation, abc.ABC):
    """Base class for unary operations on complex numbers."""

    T: ClassVar = VarConstraint("T", ComplexTypeConstr)
    complex = operand_def(T)
    result = result_def(T)
    fastmath = prop_def(FastMathFlagsAttr, default_value=FastMathFlagsAttr("none"))

    traits = traits_def(Pure())

    assembly_format = (
        "$complex (`fastmath` `` $fastmath^)? attr-dict `:` type($complex)"
    )

    def __init__(
        self,
        operand: SSAValue | Operation,
        fastmath: FastMathFlagsAttr | None = None,
    ):
        operand_ssa = SSAValue.get(operand)
        super().__init__(
            operands=[operand_ssa],
            result_types=[operand_ssa.type],
            properties={"fastmath": fastmath},
        )


class ComplexUnaryRealResultOperation(IRDLOperation, abc.ABC):
    """Base class for unary operations on complex numbers that return a float."""

    T: ClassVar = VarConstraint("T", AnyFloatConstr)

    complex = operand_def(ComplexType.constr(T))
    result = result_def(T)
    fastmath = prop_def(FastMathFlagsAttr, default_value=FastMathFlagsAttr("none"))

    traits = traits_def(Pure())

    assembly_format = (
        "$complex (`fastmath` `` $fastmath^)? attr-dict `:` type($complex)"
    )

    def __init__(
        self,
        operand: SSAValue | Operation,
        fastmath: FastMathFlagsAttr | None = None,
    ):
        operand_ssa = SSAValue.get(operand)
        element_type = cast(ComplexType, operand_ssa.type).element_type
        super().__init__(
            operands=[operand_ssa],
            result_types=[element_type],
            properties={"fastmath": fastmath},
        )

    def verify_(self):
        element_type = cast(ComplexType, self.complex.type).element_type
        if self.result.type != element_type:
            raise VerifyException(
                f"result type {self.result.type} does not match "
                f"complex element type {element_type}"
            )


class ComplexBinaryOp(IRDLOperation, abc.ABC):
    """Base class for binary operations on complex numbers."""

    T: ClassVar = VarConstraint("T", ComplexTypeConstr)
    lhs = operand_def(T)
    rhs = operand_def(T)
    result = result_def(T)
    fastmath = prop_def(FastMathFlagsAttr, default_value=FastMathFlagsAttr("none"))

    traits = traits_def(Pure())

    assembly_format = (
        "$lhs `,` $rhs (`fastmath` `` $fastmath^)? attr-dict `:` type($result)"
    )

    def __init__(
        self,
        lhs: SSAValue | Operation,
        rhs: SSAValue | Operation,
        fastmath: FastMathFlagsAttr | None = None,
    ):
        lhs_ssa = SSAValue.get(lhs)
        super().__init__(
            operands=[lhs_ssa, rhs],
            result_types=[lhs_ssa.type],
            properties={"fastmath": fastmath},
        )


class ComplexCompareOp(IRDLOperation, abc.ABC):
    """Base class for comparison operations on complex numbers."""

    T: ClassVar = VarConstraint("T", ComplexTypeConstr)
    lhs = operand_def(T)
    rhs = operand_def(T)
    result = result_def(IntegerType(1))

    traits = traits_def(Pure())

    assembly_format = "$lhs `,` $rhs attr-dict `:` type($lhs)"

    def __init__(self, lhs: SSAValue | Operation, rhs: SSAValue | Operation):
        super().__init__(operands=[lhs, rhs], result_types=[IntegerType(1)])


@irdl_op_definition
class AbsOp(ComplexUnaryRealResultOperation):
    name = "complex.abs"


@irdl_op_definition
class AddOp(ComplexBinaryOp):
    name = "complex.add"


@irdl_op_definition
class AngleOp(ComplexUnaryRealResultOperation):
    name = "complex.angle"


@irdl_op_definition
class Atan2Op(ComplexBinaryOp):
    name = "complex.atan2"


@irdl_op_definition
class BitcastOp(IRDLOperation):
    name = "complex.bitcast"
    operand = operand_def()
    result = result_def()

    traits = traits_def(Pure())

    assembly_format = "$operand attr-dict `:` type($operand) `to` type($result)"

    def __init__(self, operand: SSAValue | Operation, result_type: Attribute):
        super().__init__(operands=[operand], result_types=[result_type])


@irdl_op_definition
class ConjOp(ComplexUnaryComplexResultOperation):
    name = "complex.conj"


@irdl_op_definition
class ConstantOp(IRDLOperation, ConstantLikeInterface):
    name = "complex.constant"
    T: ClassVar = VarConstraint("T", AnyFloatConstr | base(IntegerType))
    value = prop_def(
        ArrayOfConstraint(
            RangeOf(
                AnyOf(
                    [
                        ParamAttrConstraint(IntegerAttr, (AnyAttr(), T)),
                        ParamAttrConstraint(FloatAttr, (AnyAttr(), T)),
                    ]
                )
            ).of_length(EqIntConstraint(2))
        )
    )

    # In contrast to other operations, `complex.constant` can
    # have any complex result type, not just floating point:
    complex = result_def(ComplexType.constr(T))

    traits = traits_def(Pure())

    assembly_format = "$value attr-dict `:` type($complex)"

    def __init__(self, value: ArrayAttr, result_type: ComplexType):
        super().__init__(properties={"value": value}, result_types=[result_type])

    def get_constant_value(self) -> Attribute:
        return self.value

    @staticmethod
    def from_floats(value: tuple[float, float], type: AnyFloat) -> ConstantOp:
        return ConstantOp(
            ArrayAttr([FloatAttr(value[0], type), FloatAttr(value[1], type)]),
            ComplexType(type),
        )

    @staticmethod
    def from_ints(value: tuple[int, int], type: IntegerType) -> ConstantOp:
        return ConstantOp(
            ArrayAttr([IntegerAttr(value[0], type), IntegerAttr(value[1], type)]),
            ComplexType(type),
        )


@irdl_op_definition
class CosOp(ComplexUnaryComplexResultOperation):
    name = "complex.cos"


@irdl_op_definition
class CreateOp(IRDLOperation):
    name = "complex.create"
    T: ClassVar = VarConstraint("T", AnyFloatConstr)
    real = operand_def(T)
    imaginary = operand_def(T)
    complex = result_def(ComplexType.constr(T))

    traits = traits_def(Pure())

    assembly_format = "$real `,` $imaginary attr-dict `:` type($complex)"

    def __init__(
        self,
        real: SSAValue | Operation,
        imaginary: SSAValue | Operation,
        result_type: ComplexType,
    ):
        super().__init__(operands=[real, imaginary], result_types=[result_type])


@irdl_op_definition
class DivOp(ComplexBinaryOp):
    name = "complex.div"


@irdl_op_definition
class EqualOp(ComplexCompareOp):
    name = "complex.eq"


@irdl_op_definition
class ExpOp(ComplexUnaryComplexResultOperation):
    name = "complex.exp"


@irdl_op_definition
class Expm1Op(ComplexUnaryComplexResultOperation):
    name = "complex.expm1"


@irdl_op_definition
class ImOp(ComplexUnaryRealResultOperation):
    name = "complex.im"


@irdl_op_definition
class LogOp(ComplexUnaryComplexResultOperation):
    name = "complex.log"


@irdl_op_definition
class Log1pOp(ComplexUnaryComplexResultOperation):
    name = "complex.log1p"


@irdl_op_definition
class MulOp(ComplexBinaryOp):
    name = "complex.mul"


@irdl_op_definition
class NegOp(ComplexUnaryComplexResultOperation):
    name = "complex.neg"


@irdl_op_definition
class NotEqualOp(ComplexCompareOp):
    name = "complex.neq"


@irdl_op_definition
class PowOp(ComplexBinaryOp):
    name = "complex.pow"


@irdl_op_definition
class ReOp(ComplexUnaryRealResultOperation):
    name = "complex.re"


@irdl_op_definition
class RsqrtOp(ComplexUnaryComplexResultOperation):
    name = "complex.rsqrt"


@irdl_op_definition
class SignOp(ComplexUnaryComplexResultOperation):
    name = "complex.sign"


@irdl_op_definition
class SinOp(ComplexUnaryComplexResultOperation):
    name = "complex.sin"


@irdl_op_definition
class SqrtOp(ComplexUnaryComplexResultOperation):
    name = "complex.sqrt"


@irdl_op_definition
class SubOp(ComplexBinaryOp):
    name = "complex.sub"


@irdl_op_definition
class TanOp(ComplexUnaryComplexResultOperation):
    name = "complex.tan"


@irdl_op_definition
class TanhOp(ComplexUnaryComplexResultOperation):
    name = "complex.tanh"


Complex = Dialect(
    "complex",
    [
        AbsOp,
        AddOp,
        AngleOp,
        Atan2Op,
        BitcastOp,
        ConjOp,
        ConstantOp,
        CosOp,
        CreateOp,
        DivOp,
        EqualOp,
        ExpOp,
        Expm1Op,
        ImOp,
        LogOp,
        Log1pOp,
        MulOp,
        NegOp,
        NotEqualOp,
        PowOp,
        ReOp,
        RsqrtOp,
        SignOp,
        SinOp,
        SqrtOp,
        SubOp,
        TanOp,
        TanhOp,
    ],
)



"""
This module defines an abstract base class `BlockThroughputCostModel` for estimating the
throughput of a given `Block`.

It includes a concrete implementation, `MCABlockThroughputCostModel`, that uses
the `llvm-mca` tool to perform this estimation.
"""

import os
import subprocess
from abc import ABC, abstractmethod
from shutil import which
from tempfile import NamedTemporaryFile

from xdsl.backend.assembly_printer import AssemblyPrintable, AssemblyPrinter
from xdsl.ir import Block


class BlockThroughputCostModel(ABC):
    """
    An abstract base class for a throughput cost model.

    This class provides an interface for estimating the throughput of a block
    of assembly-like operations for a specific microarchitecture.
    """

    @abstractmethod
    def estimate_throughput(self, block: Block) -> float | None:
        """
        Estimates the throughput of a block of assembly-like operations.

        Throughput is defined as the number of cycles per iteration of a block.
        The throughput value gives an indication of the performance of the block.

        Args:
            block: The block of operations to analyze.

        Returns:
            The estimated throughput as a float, or None if the estimation fails.
        """
        pass


class ExternalBlockThroughputCostModel(BlockThroughputCostModel):
    """
    An abstract base class for throughput cost models that use external command-line tools.

    This class extends `BlockThroughputCostModel` and provides common functionality
    for models that rely on an external command-line tool to produce a report
    from a source file. Subclasses must implement methods to define the command
    to run (`cmd`) and name (`tool_name`), and to parse the tool's output (`process_report`).
    """

    target: str
    """
    The target triple.
    """
    arch: str
    """
    The architecture name of the target.
    """
    src_path: str

    def __init__(self, *, target: str, arch: str):
        self.target = target
        self.arch = arch

    @classmethod
    @abstractmethod
    def tool_name(cls) -> str:
        pass

    @abstractmethod
    def cmd(self) -> list[str]:
        pass

    def produce_report(self) -> str:
        cmd = self.cmd()
        with NamedTemporaryFile(mode="w+", suffix=".s", delete=False) as report_file:
            _ = subprocess.run(
                cmd, stdout=report_file, stderr=subprocess.DEVNULL, check=True
            )
            report_file.seek(0)
            return report_file.read()

    @abstractmethod
    def process_report(self, report: str) -> float | None:
        pass

    @classmethod
    def is_installed(cls) -> bool:
        return which(cls.tool_name()) is not None

    def estimate_throughput(self, block: Block) -> float | None:
        with NamedTemporaryFile(mode="w+", delete=False, suffix=".s") as tmp_file:
            self.src_path = tmp_file.name
            printer = AssemblyPrinter(stream=tmp_file)
            for op in block.walk():
                assert isinstance(op, AssemblyPrintable), (
                    f"Block operation {op} should be an assembly instruction"
                )
                op.print_assembly(printer)

        try:
            report = self.produce_report()
            return self.process_report(report)
        finally:
            if os.path.exists(self.src_path):
                os.remove(self.src_path)


class MCABlockThroughputCostModel(ExternalBlockThroughputCostModel):
    """
    A throughput cost model that uses the `llvm-mca` tool.
    """

    @classmethod
    def tool_name(cls) -> str:
        return "llvm-mca"

    def cmd(self) -> list[str]:
        return [
            self.tool_name(),
            f"-mtriple={self.target}",
            f"-mcpu={self.arch}",
            self.src_path,
        ]

    def process_report(self, report: str) -> float | None:
        cycles = None
        iterations = None
        assert report is not None

        for l in report.splitlines():
            if l.startswith("Iterations"):
                iterations = float(l.split(":")[1])
            elif l.startswith("Total Cycles"):
                cycles = float(l.split(":")[1])

        if iterations is not None and cycles is not None:
            return cycles / iterations

        return None



from xdsl.backend.assembly_printer import reg
from xdsl.backend.register_type import RegisterType
from xdsl.irdl import irdl_attr_definition
from xdsl.utils.test_value import create_ssa_value


@irdl_attr_definition
class TestRegister(RegisterType):
    name = "test.reg"

    @classmethod
    def index_by_name(cls) -> dict[str, int]:
        return {"x0": 0, "x1": 1, "a0": 0, "a1": 1}

    @classmethod
    def infinite_register_prefix(cls):
        return "y"


def test_reg():
    x0_val = create_ssa_value(TestRegister.from_name("x0"))
    x1_val = create_ssa_value(TestRegister.from_name("x1"))
    assert reg(x0_val) == "x0"
    assert reg(x1_val) == "x1"



"""Generate the code reference pages and navigation."""

import os
from pathlib import Path

import mkdocs_gen_files
from mkdocs_gen_files.nav import Nav

nav = Nav()

root = Path(__file__).parent.parent
src = root / "xdsl"
docs_root = root / "docs"


def gen_reference():
    for path in sorted(src.rglob("*.py")):
        contents = path.read_text().strip()
        if not contents or contents.startswith("# TID 251"):
            # If this file is empty, or is an __init__.py with star imports, continue
            continue

        module_path = path.relative_to(src).with_suffix("")
        parts = tuple(module_path.parts)

        if parts[-1] == "__main__":
            continue
        elif parts[-1].startswith("_") and parts[-1] != "__init__":
            # skip private modules
            continue
        if not parts:
            continue

        doc_path = path.relative_to(src).with_suffix(".md")
        full_doc_path = Path("reference", doc_path)

        if parts[-1] == "__init__":
            parts = parts[:-1]
            if not parts:
                # skip the root __init__.py
                continue
            doc_path = doc_path.with_name("index.md")
            full_doc_path = full_doc_path.with_name("index.md")

        ident = ".".join(parts)

        nav[parts] = doc_path.as_posix()

        with mkdocs_gen_files.open(full_doc_path, "w") as fd:
            fd.write(f"::: xdsl.{ident}")

        mkdocs_gen_files.set_edit_path(full_doc_path, path.relative_to(root))


if os.environ.get("SKIP_GEN_PAGES") != "1":
    gen_reference()

# Generate an index page to empty if `gen_reference` did not run
with mkdocs_gen_files.open("reference/index.md", "w") as nav_file:
    nav_file.writelines(nav.build_literate_nav())


NEW_MARIMO_NOTEBOOK_NAMES = [
    "expressions.py",
    "eqsat.py",
    "mlir_introduction.py",
    "pdl.py",
]

NEW_MARIMO_NOTEBOOKS = [
    docs_root / "marimo" / name for name in NEW_MARIMO_NOTEBOOK_NAMES
]
"""
Notebooks expected to be run inline in mkdocs-marimo.
Some features are not supported so they have to be opted into it one by one.
"""


def gen_marimo_old():
    def create_marimo_app_url(code: str, mode: str = "read") -> str:
        from lzstring2 import LZString

        encoded_code = LZString.compress_to_encoded_URI_component(code)
        return f"https://marimo.app/#code/{encoded_code}&embed=true"

    for path in sorted((docs_root / "marimo").rglob("*.py")):
        if path in NEW_MARIMO_NOTEBOOKS:
            continue
        doc_path = path.relative_to(docs_root).with_suffix(".html")

        url = create_marimo_app_url(path.read_text())

        with mkdocs_gen_files.open(doc_path, "w") as fd:
            # Hide the header then inline the notebook
            fd.write(
                f"""\
    <iframe style="border: 0px" height="3500em" scrolling="no" width="100%" src="{url}"></iframe>
    """
            )


SYNC_XDSL_IMPORT = """\
def _():
    from xdsl.utils import marimo as xmo
    return (xmo,)
"""


def replace_xdsl_import(path: Path, destination_dir: Path):
    # Copy over the original notebook, replacing SYNC_XDSL_IMPORT with the contents of ./marimo_import_xdsl_wheel.py

    # Read the original notebook
    notebook_text = path.read_text(encoding="utf-8")

    # Read the contents of marimo_import_xdsl_wheel.py
    import_path = Path(__file__).parent / "marimo_import_xdsl_wheel.py"
    import_code = import_path.read_text(encoding="utf-8").rstrip()

    # Replace the SYNC_XDSL_IMPORT string with the import_code
    # Do not use regex; use simple string replacement
    if SYNC_XDSL_IMPORT not in notebook_text:
        raise ValueError("SYNC_XDSL_IMPORT string not found in notebook text")
    notebook_text = notebook_text.replace(SYNC_XDSL_IMPORT, import_code)

    # Write the modified notebook to the temp directory
    modified_notebook_path = destination_dir / path.name
    modified_notebook_path.write_text(notebook_text, encoding="utf-8")

    return modified_notebook_path


def gen_marimo_new_marimo():
    import subprocess
    import tempfile

    for path in NEW_MARIMO_NOTEBOOKS:
        # Create a temporary directory for marimo export
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_path = Path(temp_dir)

            modified_notebook_path = replace_xdsl_import(path, temp_path)

            # Run marimo export to temporary directory
            subprocess.run(
                [
                    "marimo",
                    "export",
                    "html-wasm",
                    # "--no-show-code",
                    "--no-sandbox",
                    str(modified_notebook_path),
                    "-o",
                    str(temp_path),
                ],
                stdout=subprocess.PIPE,
                check=True,
                encoding="utf-8",
            )

            # Copy all generated files to mkdocs using mkdocs_gen_files
            notebook_name = path.stem

            # Copy index.html
            index_html_path = temp_path / "index.html"
            if index_html_path.exists():
                with mkdocs_gen_files.open(
                    f"marimo/html/{notebook_name}/index.html", "w"
                ) as fd:
                    fd.write(index_html_path.read_text())

            # Copy assets directory if it exists
            assets_dir = temp_path / "assets"
            if assets_dir.exists():
                for asset_file in assets_dir.rglob("*"):
                    if asset_file.is_file():
                        relative_path = asset_file.relative_to(temp_path)
                        with mkdocs_gen_files.open(
                            f"marimo/html/{notebook_name}/{relative_path}", "wb"
                        ) as fd:
                            fd.write(asset_file.read_bytes())

            # Copy other static files (icons, manifests, etc.)
            for file_path in temp_path.glob("*"):
                if file_path.is_file() and file_path.name != "index.html":
                    with mkdocs_gen_files.open(
                        f"marimo/html/{notebook_name}/{file_path.name}", "wb"
                    ) as fd:
                        fd.write(file_path.read_bytes())

        # Create a markdown file that mkdocs can link to
        doc_path = path.relative_to(docs_root).with_suffix(".html")

        with mkdocs_gen_files.open(doc_path, "w") as fd:
            # Create an HTML page that redirects to the generated marimo app
            relative_path = f"html/{notebook_name}/index.html"
            fd.write(
                f"""<!DOCTYPE html>
<html>
<head>
    <meta charset="utf-8">
    <title>Redirecting to Marimo App...</title>
    <meta http-equiv="refresh" content="0; url={relative_path}">
    <style>
        body {{ font-family: Arial, sans-serif; text-align: center; padding: 50px; }}
        .loading {{ color: #666; }}
        a {{ color: #007acc; text-decoration: none; }}
        a:hover {{ text-decoration: underline; }}
    </style>
</head>
<body>
    <div class="loading">
        <p>Redirecting to interactive Marimo notebook...</p>
        <p>If you are not redirected automatically, <a href="{relative_path}">click here</a>.</p>
    </div>
</body>
</html>"""
            )


gen_marimo_old()
gen_marimo_new_marimo()

# Replace links in the marimo README
with open("docs/marimo/README.md") as rf:
    marimo_readme = rf.read()

with mkdocs_gen_files.open("marimo/index.md", "w") as fd:
    for name in NEW_MARIMO_NOTEBOOK_NAMES:
        # Replace occurrences of notebook names in NEW_MARIMO_NOTEBOOK_NAMES with
        # readonly html versions.
        marimo_readme = marimo_readme.replace(name, "html/" + name[:-3] + "/index.html")
    fd.write(marimo_readme.replace(".py", ".html"))
